//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "ConfigManager.hpp"
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#if defined WIN32
    #include <windows.h>
#else
    #include <limits.h> // Permet d'utiliser PATH_MAX, longeur max d'un path.
#endif
// Locales
#include <libintl.h>
#include <locale.h>
//------------------------------------------------------------------------------
#include "config.hpp"
#include "MyException.hpp"
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
/** Constructor */
Arguments::Arguments()
{

}
//------------------------------------------------------------------------------
/** Get the number of argument */
int Arguments::GetCount() const
{
    return m_argList.size();
}
//------------------------------------------------------------------------------

/** Get the next argument (or the first one if you call this function for the first time). */
std::string Arguments::GetNext()
{
    string arg = m_argList.front();
    m_argList.pop();
    return arg;
}
//------------------------------------------------------------------------------

/** Add an argument at the end of the list. */
void Arguments::Add(const std::string &arg)
{
    m_argList.push(arg);
}
//------------------------------------------------------------------------------

/** Clear all stored arguments */
void Arguments::Clear()
{
    while(! m_argList.empty())
    {
        m_argList.pop();
    }
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
ConfigNode::ConfigNode() : m_node(NULL)
{

}
//------------------------------------------------------------------------------
// Constructor
/** Construct ConfigMode from a ticpp Element */
ConfigNode::ConfigNode(ticpp::Element *node) : m_node(node)
{

}
//------------------------------------------------------------------------------

// Get
/** Get the string corresponding to key in this node
 */
std::string ConfigNode::GetString(const std::string &key) const
{
    try
    {
        return m_node->FirstChildElement(key)->GetAttribute("value");
    }
    catch(...)
    {
        cerr << "Can't get string '" << key << "'." << endl;
        return "";
    }
}
//------------------------------------------------------------------------------

/** Get the number corresponding to key in this node
 */
float ConfigNode::GetNumber(const std::string &key) const
{
    try
    {
        return m_node->FirstChildElement(key)->GetAttribute<float>("value");
    }
    catch(...)
    {
        cerr << "Can't get number '" << key << "'." << endl;
        return 0;
    }
}
//------------------------------------------------------------------------------
bool ConfigNode::ChildExists(const std::string &key) const
{
    return (m_node->FirstChildElement(key, false) != NULL);
}
//------------------------------------------------------------------------------

/** Get the ConfigNode corresponding to key, if it exists, otherwise, create a child ConfigNode named "key" in this ConfigNode and return it */
ConfigNode ConfigNode::AccessNode(const std::string &key)
{
    ticpp::Element *elem;
    if((elem = m_node->FirstChildElement(key, false)) == NULL) // Doesn't exist
    {
        elem = new ticpp::Element(key);
        m_node->LinkEndChild(elem);
    }
    else if(elem->HasAttribute("value")) // Should nove have attribute value
    {
        elem->RemoveAttribute("value");
    }

    return ConfigNode(elem);
}
//------------------------------------------------------------------------------

// Set
/** Set / create a string with key */
void ConfigNode::SetValue(const std::string &key, const std::string &value)
{
    ticpp::Element *elem;
    if((elem = m_node->FirstChildElement(key, false)) == NULL)
    {
        elem = new ticpp::Element(key);
        m_node->LinkEndChild(elem);
    }

    elem->SetAttribute("value", value);
}
//------------------------------------------------------------------------------

/** Set / create a number with key */
void ConfigNode::SetValue(const std::string &key, float value)
{
    ticpp::Element *elem;
    if((elem = m_node->FirstChildElement(key, false)) == NULL)
    {
        elem = new ticpp::Element(key);
        m_node->LinkEndChild(elem);
    }

    elem->SetAttribute("value", value);
}
//------------------------------------------------------------------------------
/** Get a number and write default value if it doesn't exist */
float ConfigNode::GetNumberDefault(const std::string &key, float defaultVal)
{
    try
    {
        return m_node->FirstChildElement(key)->GetAttribute<float>("value");
    }
    catch(...)
    {
        SetValue(key, defaultVal);
        return defaultVal;
    }
}
//------------------------------------------------------------------------------

// Remove
/** Remove the child named key */
void ConfigNode::Remove(const std::string &key)
{
    ticpp::Element *elem;
    if((elem = m_node->FirstChildElement(key, false)) != NULL)
    {
        m_node->RemoveChild(elem);
    }
}
//------------------------------------------------------------------------------

/** Clear the ConfigNode (delete all childs) */
void ConfigNode::Clear()
{
    m_node->Clear();
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
ConfigManager::ConfigManager() : m_currentMode(NULL)
{
    /** Load useful paths. Should be called first */
    LoadStandardPaths();
}
//------------------------------------------------------------------------------
ConfigManager::~ConfigManager()
{
    SaveXmlConfig(m_globalConfigDoc, m_globalConfig);
    SaveXmlConfig(m_modeConfigDoc, m_modeConfig);
}
//------------------------------------------------------------------------------
void ConfigManager::LoadAll()
{
    LoadModeList();

    LoadXmlConfig(m_userConfigPath + "/mainconfig.xml", m_globalConfigDoc, m_globalConfig);
}
//------------------------------------------------------------------------------
void ConfigManager::LoadStandardPaths()
{
    std::string exeName, exeDir, exeTotalPath;

    #if defined (WIN32)
    // Data dir
    char exepath[4096] = {0};
    GetModuleFileName(NULL, exepath, sizeof(exepath));  /* get fullpath of the service */

    exeTotalPath = exepath;
    m_dataPath = exeTotalPath.substr(0,exeTotalPath.rfind('\\') + 1) + std::string("data");

    // User config dir
    m_userConfigPath = getenv("APPDATA");
    m_userConfigPath += std::string("/.") + PACKAGE;
    mkdir(m_userConfigPath.c_str());
    #else
    // Data dir
    char exepath[PATH_MAX] = {0};
    char procid[PATH_MAX] = {0};
    char exelink[PATH_MAX] = {0};

    sprintf(procid, "%u", getpid());

    strcpy(exelink, "/proc/");
    strcat(exelink, procid);
    strcat(exelink, "/exe");

    int result = readlink(exelink, exepath, sizeof(exepath));
    if ( result != -1 )
    {
        exepath[result] = '\0'; // readlink() doesn't NUL-terminate the buffer
    }

    exeTotalPath = exepath;

    exeName = exeTotalPath.substr(exeTotalPath.rfind('/') + 1);
    exeDir = exeTotalPath.substr(0,exeTotalPath.rfind('/') + 1);

    if(exeDir.substr(exeDir.size() - 4) == std::string("bin/") || exeDir.substr(exeDir.size() - 6) == std::string("games/"))
    {
        m_dataPath = exeDir + std::string("../share/") + exeName + std::string("/data");
    }
    else // Local
    {
        m_dataPath = exeDir + std::string("data");
    }

    // User config dir
    std::string userDir = getenv("HOME");
    m_userConfigPath = userDir + std::string("/.") + PACKAGE;
    mkdir(m_userConfigPath.c_str(), 0777);
    #endif

    m_dataCommonPath = m_dataPath + "/common";
}
//------------------------------------------------------------------------------
/** Select a mode to use. It will lunch main.lua. */
void ConfigManager::LoadMode(const std::string &idName)
{
    std::map<std::string, ModeInfo>::iterator it = m_modList.find(idName);
    if(it == m_modList.end())
    {
        MY_THROW("Mode '" << idName << "' not found !");
    }

    // Config
    SaveXmlConfig(m_modeConfigDoc, m_modeConfig);
    LoadXmlConfig(m_userConfigPath + "/mode_" + idName + ".xml", m_modeConfigDoc, m_modeConfig);

    // Info
    m_currentMode = &(it->second);

    // Chdir
    chdir(m_currentMode->path.c_str());
}
//------------------------------------------------------------------------------

/** Get information about mode idName */
ModeInfo ConfigManager::GetModeInfo(const std::string &idName)
{
    return m_modList.at(idName);
}
//------------------------------------------------------------------------------

/** Get information about mode number id */
ModeInfo ConfigManager::GetModeInfo(int id)
{
    if(id >= 0 && id < GetModeCount())
    {
        std::map<string, ModeInfo>::const_iterator it = m_modList.begin();
        for(int i = 0; i < id; i++)
        {
            it++;
        }

        return it->second;
    }
    else
    {
        MY_THROW("GetModeInfo: id out of range !");
    }
}
//------------------------------------------------------------------------------

/** Get the current mode id */
std::string ConfigManager::GetCurrentModeId() const
{
    if(m_currentMode)
    {
        return m_currentMode->idName;
    }
    else
    {
        return "None";
    }
}
//------------------------------------------------------------------------------

/** Get the root ConfigNode for the current mode */
ConfigNode ConfigManager::GetRootConfigNode()
{
    return m_modeConfig;
}
//------------------------------------------------------------------------------
/** Get the global ConfigNode */
ConfigNode ConfigManager::GetGlobalConfigNode()
{
    return m_globalConfig;
}
//------------------------------------------------------------------------------

/** Return the number of mode */
int ConfigManager::GetModeCount() const
{
    return m_modList.size();
}
//------------------------------------------------------------------------------
/** Create a new RenderWindow, wich should be deleted by the programm */
sf::RenderWindow *ConfigManager::CreateRenderWindow()
{
    ConfigNode conf = GetGlobalConfigNode().AccessNode("graphicConf");

    bool isFullscreen = conf.GetNumberDefault("isFullscreen", true);
    bool isVsyn = conf.GetNumberDefault("isVsyn", true);
    int width = conf.GetNumberDefault("width", -1);
    int height = conf.GetNumberDefault("height", -1);
    int antialiasingLevel = conf.GetNumberDefault("antialiasingLevel", 2);
    string windowName = PACKAGE_NAME;
    if(m_currentMode)
    {
         windowName += " - " + m_currentMode->fullName;
    }

    sf::VideoMode mode(width, height);

    if(isFullscreen)
    {
        if(height <= 0 || width <= 0 || !mode.IsValid())
        {
            mode = sf::VideoMode::GetDesktopMode();
        }
    }
    else
    {
		sf::VideoMode maxMode = sf::VideoMode::GetDesktopMode();
        if(height <= 0 || width <= 0 || height > int(maxMode.Height) || width > int(maxMode.Width))
        {
            mode = sf::VideoMode(maxMode.Width * (2.f/3.f), maxMode.Height * (2.f/3.f));
        }
    }

    unsigned long flag = sf::Style::Resize | sf::Style::Close;
    if(isFullscreen)
    {
        flag |= sf::Style::Fullscreen;
    }

    // Create the main window
    sf::RenderWindow *App = new sf::RenderWindow(mode, windowName, flag, sf::WindowSettings(24, 8, antialiasingLevel));
    if(!App)
    {
        MY_THROW("Can't create render window !");
    }

    App->UseVerticalSync(isVsyn);
    App->EnableKeyRepeat(false);

    // Icon
    if(m_iconImg.LoadFromFile(GetCommonDataPath() + "/icon.png"))
    {
        App->SetIcon(m_iconImg.GetWidth(), m_iconImg.GetHeight(), m_iconImg.GetPixelsPtr());
    }
    else
    {
        cerr << "Can't find icon '" << GetCommonDataPath() << "/icon.png' !" << endl;
    }

    return App;
}
//------------------------------------------------------------------------------
void ConfigManager::LoadXmlConfig(const std::string &filename, ticpp::Document &doc, ConfigNode &rootNode)
{
    try
    {
        doc.LoadFile(filename);
        ticpp::Element *root = doc.FirstChildElement("modeConfig_file");
        if(!root->GetAttribute<int>("version") == 1)
        {
            MY_THROW("Bad version !");
        }

        rootNode.SetXmlNode(root);
    }
    catch(std::exception &e)
    {
        cerr << "Failed to load config '" << filename << "' !" << endl;
        cerr << "Cause: " << e.what() << endl;
        cerr << "Doc will be reseted" << endl;

        doc = ticpp::Document(filename);

        doc.LinkEndChild(new ticpp::Declaration("1.0","UTF-8","no"));
        ticpp::Element *root = new ticpp::Element("modeConfig_file");
        doc.LinkEndChild(root);
        root->SetAttribute("version", 1);

        rootNode.SetXmlNode(root);
    }
}
//------------------------------------------------------------------------------
void ConfigManager::SaveXmlConfig(ticpp::Document &doc, ConfigNode &node)
{
    if(node.GetXmlNode() != NULL)
    {
        doc.SaveFile();
    }
}
//------------------------------------------------------------------------------
void ConfigManager::LoadModeList()
{
    m_modList.clear();

    vector<string> pathList = GetConfigDirs();
    for(vector<string>::iterator it = pathList.begin(); it != pathList.end(); it++)
    {
        string lookupPath = *it + "/modes";
        vector<string> contentList = GetDirsInPath(lookupPath);
        for(vector<string>::iterator itIdName = contentList.begin(); itIdName != contentList.end(); itIdName++)
        {
            try
            {
                string modePath = lookupPath + "/" + *itIdName;
                string filename = modePath + "/info.xml";

                ticpp::Document myDoc(filename);
                myDoc.LoadFile();
                ticpp::Element *root = myDoc.FirstChildElement("modeinfo_file");
                if(!root->GetAttribute<int>("version") == 1)
                {
                    MY_THROW("Bad version !");
                }

                const ConfigNode infoNode(root);
                ModeInfo &info = m_modList[*itIdName];
                info.fullName = infoNode.GetString("fullName");
                info.idName = *itIdName;
                info.auther = infoNode.GetString("auther");
                info.path = modePath;
                info.description = infoNode.GetString("description");
                info.version = infoNode.GetString("version");
                info.screenshot = infoNode.GetString("screenshot");
            }
            catch(std::exception &e)
            {
                cerr << "Failed to load mode info from '" << *itIdName << "' !" << endl;
                cerr << "Cause: " << e.what() << endl;
            }
        }
    }
}
//------------------------------------------------------------------------------
std::vector<std::string> ConfigManager::GetDirsInPath(const std::string &path)
{
    std::vector<string> list;

    DIR *dir = NULL;

    dir = opendir(path.c_str());
    if(dir == NULL)
    {
        cerr << "Can't open dir '" << path << "' !" << endl;
        return list; // Return empty list
    }

    for(dirent *myDirent = readdir(dir); myDirent != NULL; myDirent = readdir(dir))
    {
        string name = myDirent->d_name;
        if(name != "." && name != "..")
        {
            DIR *curDir = NULL;
            string currentPath = path + "/" + name;
            curDir = opendir(currentPath.c_str());
            if(curDir != NULL)
            {
                closedir(curDir);
                list.push_back(name);
            }
        }
    }

    closedir(dir);

    return list;
}
//------------------------------------------------------------------------------
std::vector<std::string> ConfigManager::GetConfigDirs()
{
    std::vector<std::string> list;
    list.push_back(m_dataPath);
    list.push_back(m_userConfigPath);
    return list;
}
//------------------------------------------------------------------------------
/** Get the path of the common data dir (data/common) */
std::string ConfigManager::GetCommonDataPath()
{
    return m_dataCommonPath;
}
//------------------------------------------------------------------------------
/** Get the path of the mode dir (data/modes/[modeName]) */
std::string ConfigManager::GetModePath()
{
    return m_currentMode->path;
}
//------------------------------------------------------------------------------
Arguments &ConfigManager::GetArguments()
{
    return m_arguments;
}
//------------------------------------------------------------------------------
/** Init the locales */
void ConfigManager::InitLocales(const std::string &path)
{
    if(setlocale(LC_ALL, "") == NULL)
    {
        MY_THROW("ConfigManager::InitLocales(): Can't init locales !")
    }

    if(bindtextdomain("bottle2d", path.c_str()) == NULL)
    {
        MY_THROW("ConfigManager::InitLocales(): Can't init locales !")
    }

    if(textdomain( "bottle2d") == NULL)
    {
        MY_THROW("ConfigManager::InitLocales(): Can't init locales !")
    }
}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
