//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include <vector>
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
//------------------------------------------------------------------------------
#include "Utils.hpp"
#include "ConfigManager.hpp"
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
#ifndef GLOBAL_HPP
#define GLOBAL_HPP
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
struct lua_State;
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
class Screen;
class ScreenConfig;
class Console;
class ButtonEventHandler;
class Group;
//------------------------------------------------------------------------------

/// LUA_HEADER_COPY_ON ///
/** \\file header.hpp
    \\brief All global functions and classes.
*/
/// LUA_HEADER_COPY_OFF ///
namespace Global
{
    //------------------------------------------------------------------------------

    class Engine
    {
    public:
        Engine(int argc, const char* argv[]);
        virtual ~Engine();

    private:
        Engine(const Engine &en);
        Engine();
    };

    // Render Window
    sf::RenderWindow &GetApp();

    // Lua
    void LuaRunScript(const std::string &filename);
    lua_State *GetLuaState();

    // Screen
    void AddScreen(Screen *screen);
    void RemoveScreen(Screen *screen);

    /// LUA_EXPORT_ON() ///
    // Lua
    /** Select a new screen from its name */
    void ChangeScreen(const std::string &name);

    /** Select a new screen from the associated screen object */
    void ChangeScreen(Screen *screen);

    /** Create a new screen named [name] and return a handler to it */
    Screen *CreateScreen(const std::string &name);

    /** Create a new screen named [name] and return a handler to it. This function allow you to set a ScreenConfig to this screen. */
    Screen *CreateScreen(const std::string &name, const ScreenConfig &screenConfig);

    // Paths
    /** Get the path of the common data dir (data/common) */
    std::string GetCommonDataPath();

    /** Get the path of the mode dir (data/modes/[modeName]) */
    std::string GetModePath();

    /** Get the arguments.
      * Note: if you want to pass argument to a mode, you just have to add arguments into this object.
     */

    // Info
    /** Get the version of the engine */
    std::string GetEngineVersion();

    // Render Window
    /** Recreate a render window, after changing settings */
    void RequestRecreateRenderWindow();

    // Other
    /** Get the current frime time\\n
      * Note: 1/GetFrameTime() gives you the frame per second (FPS)
      */
    float GetFrameTime();

    // Other
    /** Exit the engine */
    void Quit();

    // RatioEdit
    /** Convert a ratio (between 0 and 1) to make it grow faster and faster (from 0 to 0.5) and then slower and slower (from 0.5 to 1).\\n
      * Note: This function ensure that ConvertRatioCosAccelerationDeceleration(0) = 0 and ConvertRatioCosAccelerationDeceleration(1) = 1.
      */
    float ConvertRatioCosAccelerationDeceleration(float ratio);

    /** Convert a ratio (between 0 and 1) to make it grow faster and faster (exponential type).\\n
      * You have to give an acceletation coefficient (> 0).\\n
      * Note: This function ensure that ConvertRatioAcceleration(0) = 0 and ConvertRatioAcceleration(1) = 1.
      */
    float ConvertRatioAcceleration(float ratio, float coefficient);

    /** Convert a ratio (between 0 and 1) to make it grow slower and slower (exponential type).\\n
      * You have to give an acceletation coefficient (> 0).\\n
      * Note: This function ensure that ConvertRatioDeceleration(0) = 0 and ConvertRatioDeceleration(1) = 1.
      */
    float ConvertRatioDeceleration(float ratio, float coefficient);

    // Input
    /** Return the current mouse position */
    Vector2D GetMousePosition();

    // Config
    /** Return the configuration node attached to the currently runing mode (your mode) */
    ConfigNode GetModeConf();

    /** Return the informations (ModeInfo) attached to the currently runing mode (your mode) */
    ModeInfo GetModeInfo();

    /** Laod the mode idName (idName is the name of the mode directory, in data/modes/ or the user config directory */
    void LoadMode(const std::string &idName);

    /** Get the ConfigManager of the engine */
    ConfigManager &GetConfigManager();

    /** Force the engine to draw a frame. Usefull for progress bar, by example */
    void ForceFrame();

    /** Show or hide the information console */
    void ShowConsole(bool isVisible);

    /** Return true if the information console is visible, false if not */
    bool IsConsoleVisible();

    /** Return true if the key \\a keyName is pressed, false otherwise. See \\ref page_keyList for the list of keys. */
    bool IsKeyPressed(const std::string &keyName);

    /** Return the real size of the window (resolution). This function should not be used unless you want to know the real window ratio.\\n
      * See ScreenConfig.
      */
    Vector2D GetRealWindowSize();

    /** Set if the group centers and the physic objects should be drawn (helps for debuging). */
    void SetShowPhysicObjects(bool isVisible);

    /** Return true if the group centers and the physic objects should be drawn, else otherwise. */
    bool IsShowPhysicObjects();

    /** Translation function. Use Gettext to translate your string. See ConfigManager::InitLocales(). */
    std::string _(const std::string &str);

    /** Return the type name (string) of the given type id */
    std::string GetTypeName(int typeId);

    /** Get the child tree (for debuging purpose) of the currently running screen. */
    std::string GetChildTree();

    /** Get the avaible video modes (return a table of Vector2D).\\n
     *  Note: the first element of the table is Vector2D(-1, -1), it means the mode is automaticaly chosen a the start of the engine.
     */
    luabind::object GetAvaibleVideoModes();

    /** Get the arguments given to bottle2d when executed (return a table of string). */
    luabind::object GetCmdArguments();
    /// LUA_EXPORT_OFF ///

    /** Get the arguments given to bottle2d when executed (return a table of std::string). */
    std::vector<std::string> GetStdCmdArguments();

    /** Return a reference to the information console */
    Console &GetConsole();

    // Other
    void Run();

    // Unload a mode and destroy its screen
    void Clear();

    // Lua C API
    // Echo
    /** Print a message to the engine's console, and to standard output (stdout) */
    int printConsole(lua_State* L);

    /** Print an error message to the engine's console, and to standard error output (stderr) */
    int printErrorConsole(lua_State* L);


    namespace Priv
    {
        // FIXME : Seulement en debug
        void CheckEngine() throw();

        void LoadLuaState();

        // Hotkeys
        const sf::Key::Code HOTKEY_KEY = sf::Key::LControl; // LControl need to be pressed.

        void OnHotkeyQuit(const std::string &keyName, bool pressed, ButtonEventHandler *obj, const luabind::object &luaFunction);
        void OnHotkeySwitchFullscreen(const std::string &keyName, bool pressed, ButtonEventHandler *obj, const luabind::object &luaFunction);
        void OnHotkeyShowConsole(const std::string &keyName, bool pressed, ButtonEventHandler *obj, const luabind::object &luaFunction);
        void OnHotkeyMenu(const std::string &keyName, bool pressed, ButtonEventHandler *obj, const luabind::object &luaFunction);
        void OnHotkeyShowPhysicObj(const std::string &keyName, bool pressed, ButtonEventHandler *obj, const luabind::object &luaFunction);

        /** Register hotkeys events */
        void AddHotKeysEvents(Screen *screen);

        /** Do load a mode */
        void DoLoadMode();

        /** Get the child tree (for debuging purpose) */
        void GetChildTreeRecursive(std::string &text, int level, Group *group);

        /** On lua panic */
        int OnLuaPanic(lua_State *state);

        /** Print the parameters to the given ostream */
        int printParameters(lua_State* L, std::ostream &stream);

        /** Create a render window */
        void CreateRenderWindow();
    } // namespace priv
} // namespace Global

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
} // namespace En
#endif // GLOBAL_HPP
