//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef PARTICULESOURCE_HPP
#define PARTICULESOURCE_HPP
//------------------------------------------------------------------------------
#include "Drawable.hpp"
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
#include <Core/SPK_DEF.h>
namespace SPK
{
    class Model;
    class Group;
    class Emitter;
    namespace SFML
    {
        class SFMLSystem;
    }
}
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A particle model defines the variable of the particle. The more numerous the variable are, the bigger memory size will be used by the ParticleSource. */
/// LUA_HEADER_COPY_OFF ///
//------------------------------------------------------------------------------
class ParticleModel
{
public:
    /// LUA_EXPORT_ON(ParticleModel) ///
    /** Constructor */
    ParticleModel();

    /** Constructor */
    ParticleModel(float lifeTimeMin, float lifeTimeMax);

    /** Add a fixed value for parameter "name".\\n
      * Note: Name can be either "red", "green", "blue" or "alpha".
      */
    void AddFixedParam(const std::string &name, float value);

    /** Add a mutable value for parameter "name". This value will change from start to end during the particle life time.\\n
      * Note: Name can be either "red", "green", "blue" or "alpha".
      */
    void AddMutableParam(const std::string &name, float begin, float end);

    /** Add a random value for parameter "name". For each particle a value will be chosen between min and max.\\n
      * Note: Name can be either "red", "green", "blue" or "alpha".
      */
    void AddRandomParam(const std::string &name, float min, float max);

    /** Add a random and mutable value for parameter "name". For each particle a value will be chosen between begin_min and begin_max. At end of life, the value will be between end_min and end_max.\\n
      * Note: Name can be either "red", "green", "blue" or "alpha".
      */
    void AddMutableRandomParam(const std::string &name, float begin_min, float begin_max, float end_min, float end_max);

    /** Set the particle life time. A random value will be chosen between min and max for each particle. */
    void SetLifeTime(float min, float max);
    /// LUA_EXPORT_OFF ///

    SPK::Model *CreateSparkModel() const;
protected:
    static SPK::ModelParam GetParamId(const std::string &name);

protected:
    typedef std::map<SPK::ModelParam, std::pair<int, float[4]> > stockMap;
    stockMap m_paramList;
    float m_lifeTimeMin, m_lifeTimeMax;

    enum
    {
        paramFixed, paramMutable, paramRandom, paramMutableRandom
    };
};
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A particle emitter defines how the particle will be emitted (their direction and emission zone). */
/// LUA_HEADER_COPY_OFF ///
class ParticleEmitter
{
public:
    /// LUA_EXPORT_ON(ParticleEmitter) ///
    /** Default contructor */
    ParticleEmitter();

    /** Set throw angles */
    void SetAngles(float angleA, float angleB);

    /** Set direction (doesn't need to be normalised) */
    void SetDirection(const Vector2D &direction);

    /** Make the emitter as a line */
    void MakeLine(const Vector2D &pt1, const Vector2D &pt2);

    /** Make the emitter as a disk */
    void MakeDisk(const Vector2D &center, float radius);

    /** Make the emitter as a circle */
    void MakeCircle(const Vector2D &center, float radius);

    /** Make the emitter as a point */
    void MakePoint(const Vector2D &position);
    /// LUA_EXPORT_OFF ///

    SPK::Emitter *CreateEmitter() const;

protected:
    // Types
    enum
    {
        TypeLine,
        TypeDisk,
        TypeCircle,
        TypePoint
    };

    // Zone
    int m_type;
    Vector2D m_pt1, m_pt2;
    float m_radius;

    // Emitter
    float m_angleA, m_angleB;
    Vector2D m_direction;
};
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

/// LUA_HEADER_COPY_ON ///
/** A particle source is used to send particle and create beautiful effects. It can emit only one type of particle.
  * See Group::CreateParticleSource to create a particle source object.
  */
/// LUA_HEADER_COPY_OFF ///
class ParticleSource : public Drawable
{
public:
    ParticleSource(Group *parent, const std::string &name, const ParticleModel &model, int maxParticleNumber);
    virtual ~ParticleSource();
    // Copy
    ParticleSource(const ParticleSource &source);

    //static ParticleSource *Create(Group *parent, const std::string &name, const ParticleModel &model) {return new ParticleSource(parent, name);}

    /// LUA_EXPORT_ON(ParticleSource, Drawable) ///
    // Clone & Create
    /** Create a clone of the object and return an handler to the clone. */
    virtual ParticleSource *Clone() const {return CloneToGroup(NULL);}

    /** Create a clone of the object, change the parent and return an handler to the clone. */
    virtual ParticleSource *CloneToGroup(Group *newParent) const {ParticleSource *obj = new ParticleSource(*this); obj->RegisterClone(newParent); return obj;}

    /** Access flow (see SetFlow) */
    //LUA// luaProperty(int, flow, GetFlow, SetFlow);

    /** Set particle flow (number of particle emitted per second).\\n
      * Value -1 means unlimited.\\n
      * Note: You can't set Flow and tank to -1 (infinite).
      */
    void SetFlow(int flow);

    /** Get particle flow */
    int GetFlow() const;

    /** Access tank (see SetTank) */
    //LUA// luaProperty(int, tank, GetTank, SetTank);

    /** Set particle tank (the number of particle that the source can emit).\\n
      * Value -1 means unlimited. When tank is empty, you can add new particle by calling SetTank()\\n
      * Note: You can't set Flow and tank to -1 (infinite).
      */
    void SetTank(int tank);

    /** Get particle tank */
    int GetTank() const;

    /** Access particle gravity */
    //LUA// luaProperty(Vector2D, gravity, GetGravity, SetGravity);

    /** Set particle gravity */
    void SetGravity(const Vector2D &force);

    /** Get gravity */
    Vector2D GetGravity() const;

    /** Access particle friction */
    //LUA// luaProperty(float, friction, GetGravity, GetFriction);

    /** Set friction */
    void SetFriction(float friction);

    /** Get friction */
    float GetFriction() const;

    /** Set emission force */
    void SetForce(float min, float max);

    /** Access minimum force */
    //LUA// luaProperty(float, minForce, GetMinForce, SetMinForce);

    /** Get min emission force */
    float GetMinForce() const;

    /** Set min emission force */
    void SetMinForce(float minForce) {SetForce(minForce, GetMaxForce());}

    /** Access maximum force */
    //LUA// luaProperty(float, maxForce, GetMaxForce, SetMaxForce);

    /** Get max emission force */
    float GetMaxForce() const;

    /** Set max emission force */
    void SetMaxForce(float maxForce) {SetForce(GetMinForce(), maxForce);}

    /** Set throw angle of the emitter (0 : only in one direction, 2*pi : ain every direction) */
    void SetEmitterAngle(float angle);

    /** Get the throw angle of the emitter (0 : only in one direction, 2*pi : ain every direction) */
    float GetEmitterAngle() const;

    /** Set direction of the emitter (doesn't need to be normalised) */
    void SetEmitterDirection(const Vector2D &direction);

    /** Set direction of the emitter (doesn't need to be normalised) */
    Vector2D GetEmitterDirection() const;

    /** Set the particle emitter. Warning: It will remove most of the paramaters registered in the ParticleSource. */
    void SetEmitter(const ParticleEmitter &emitter);

    /** Make particle rendered as lines */
    void MakeRenderLines(float length, float width);

    /** Make particle rendered as points */
    void MakeRenderPoints(float pointSize);

    /** Make particle rendered as images */
    void MakeRenderImages(Texture *img, const Vector2D &size, bool blendAdd);

    /** Access auto destroy mode (see SetAutoDestroy) */
    //LUA// luaProperty(bool, autoDestroy, IsAutoDestroy, SetAutoDestroy);

    /** Set auto destroy mode : when tank is 0, the ParticleSource should be automatically destroyed. Default is deactivated. */
    void SetAutoDestroy(bool autoDestroy);

    /** Check if auto destroy mode is activated */
    bool IsAutoDestroy() const;

    /** Conpute particle position as if \\a time seconds were elapsed. This is usefull if you to make the user think that the particle system was already running for a long time.\\n
      * \\param time Time which is supposed to be elapsed. Usualy, it should be \\a maxLifeTime.
      */
    void ComputeParticles(float time);

    /** Acces the event function (\\a fnName should be an lua function or nil for nothing) that is called when the tank is empty. */
    //LUA// luaProperty(luabind::object, onTankEmptyEventFunction, GetOnTankEmptyEventFunction, SetOnTankEmptyEventFunction);
    /** Set the event function (\\a fnName should be an existing lua function, or nil for nothing) that is called when the tank is empty.\\n
      * Function prototype should be : function fnName(thisParticleSource : ParticleSource)
      */
    void SetOnTankEmptyEventFunction(const luabind::object &luaFunction);
    /** Get the event function that is called when the tank is empty. */
    luabind::object GetOnTankEmptyEventFunction() const;

    /// LUA_EXPORT_OFF ///
    // No Lua
    // Mvt
    virtual void ProceedMvt(float time);

    // Type
    ADD_GET_TYPE_FN(OBJ_TYPE_DRAWABLE_PARTICULE_SRC)

    // Render
    virtual void Render(sf::RenderTarget &App);
protected:
    SPK::Group *m_sparkGroup;
    SPK::SFML::SFMLSystem *m_sparkSystem;

    bool m_isAutoDestroy;

    luabind::object m_luaFunction;
};
//------------------------------------------------------------------------------
} // namespace En

#endif // PARTICULESOURCE_HPP
