//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef PSEUDO_EVENT_HPP
#define PSEUDO_EVENT_HPP
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
#include "Event.hpp"

#include <string>
//------------------------------------------------------------------------------
namespace En
{

class Screen;
class PhysicObject;
//------------------------------------------------------------------------------
class PseudoEventHandler : public EventHandlerBase
{
public:
    PseudoEventHandler(const luabind::object &luaFunction);
    virtual bool Match(const Screen *screen, const sf::RenderWindow &App) = 0;
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** Timer event, to send event at regular interval, see Screen::RegisterTimer to register a timer event. */
/// LUA_HEADER_COPY_OFF ///
class Timer : public PseudoEventHandler
{
public:
    Timer(const luabind::object &luaFunction, float interval, bool isRunning = true);
    virtual bool Match(const Screen *screen, const sf::RenderWindow &App);

    /// LUA_EXPORT_ON(Timer, EventHandlerBase) ///
    // Lua
    // Interval
    /** Access the timer interval (time between two events). */
    //LUA// luaProperty(float, interval, GetInterval, SetInterval);
    /** Get the timer interval (time between two events). */
    float GetInterval() const {return m_interval;}
    /** Set the timer interval (time between two events). */
    void SetInterval(float interval) {m_interval = interval;}

    // Start / Stop
    /** Start the timer. */
    void Start() {m_currentTime = 0.f; m_isRunning = true;}
    /** Stop / pause the timer. */
    void Stop() {m_isRunning = false;}
    /** Return true if the timer is running, false overwise. */
    bool IsRunning() const {return m_isRunning;}
    /// LUA_EXPORT_OFF ///

protected:
    void SendEvent(Timer *obj, const luabind::object &luaFunction);

    float m_currentTime;
    float m_interval;
    bool m_isRunning;
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** Chronometer event, to send event on each frame while the time is not elapsed, see Screen::RegisterChronometer to register a chronometer event. */
/// LUA_HEADER_COPY_OFF ///
class Chronometer : public PseudoEventHandler
{
public:
    Chronometer(const luabind::object &luaFunction, float totalTime, bool isRunning = true);
    virtual bool Match(const Screen *screen, const sf::RenderWindow &App);

    /// LUA_EXPORT_ON(Chronometer, EventHandlerBase) ///
    // Lua
    // Interval
    /** Access the total time that the chronometer should count (when total time is elapsed, ratio is 1). */
    //LUA// luaProperty(float, totalTime, GetTotalTime, SetTotalTime);
    /** Get the total time that the chronometer should count (when total time is elapsed, ratio is 1). */
    float GetTotalTime() const {return m_totalTime;}
    /** Set the total time that the chronometer should count (when total time is elapsed, ratio is 1). */
    void SetTotalTime(float totalTime) {m_totalTime = totalTime;}


    /** Access the current time ratio(elapsed time / total time). */
    //LUA// luaProperty(float, ratio, GetRatio, SetRatio);
    /** Get the current time ratio(elapsed time / total time). */
    float GetRatio() const {return m_currentTime / m_totalTime;}
    /** Force the chronometer to force its current ratio */
    void SetRatio(float newRatio);


    /** Restart the chronometer (set ratio and elapsed time to 0). */
    void Restart() {Start(); m_currentTime = 0;}

    // Start / Stop
    /** Start the chronometer. */
    void Start() {m_isRunning = true;}
    /** Stop / pause the chronometer. */
    void Stop() {m_isRunning = false;}
    /** Return true if the chronometer is running, false overwise. */
    bool IsRunning() const {return m_isRunning;}
    /// LUA_EXPORT_OFF ///

protected:
    static void SendEvent(float ratio, Chronometer *obj, const luabind::object &luaFunction);

    float m_totalTime;
    bool m_isRunning;

    float m_currentTime;
};
//------------------------------------------------------------------------------
class FrameEvent : public PseudoEventHandler
{
public:
    FrameEvent(const luabind::object &luaFunction);
    virtual bool Match(const Screen *screen, const sf::RenderWindow &App);

protected:
    static void SendEvent(float frameTime, FrameEvent *obj, const luabind::object &luaFunction);
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** Mouse enter / leave physic object event, see Screen::RegisterMouseZoneEvent to register a MouseZone event. */
/// LUA_HEADER_COPY_OFF ///
class MouseZoneEvent : public PseudoEventHandler
{
public:
    MouseZoneEvent(const luabind::object &luaFunction, PhysicObject *zone);
    virtual bool Match(const Screen *screen, const sf::RenderWindow &App);

    /// LUA_EXPORT_ON(MouseZoneEvent, EventHandlerBase) ///
    // Lua
    // Interval
    /** Access the zone (a Physic object) used to detect mouse entering / exiting */
    //LUA// luaProperty(float, zone, GetZone, SetZone);
    /** Get the zone (a Physic object) used to detect mouse entering / exiting */
    PhysicObject *GetZone() const {return m_zone;}
    /** Set the zone (a Physic object) used to detect mouse entering / exiting */
    void SetZone(PhysicObject *zone) {m_zone = zone;}

    /** Return true if the mouse is currently inside the zone, else overwise */
    bool IsInside() const {return m_isIn;}
    /// LUA_EXPORT_OFF ///

protected:
    static void SendEvent(MouseZoneEvent *obj, const luabind::object &luaFunction);

    bool m_isIn;
    PhysicObject *m_zone;
};
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // PSEUDO_EVENT_HPP
