//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef TEXT_HPP
#define TEXT_HPP

#include <SFML/Graphics.hpp>

#include "Drawable.hpp"
//------------------------------------------------------------------------------
namespace En
{
/// LUA_HEADER_COPY_ON ///
/** A Text is used to render text. Font, style and size can be changed. It is also posible to enable auto line return. It should be created using Group::CreateText. */
/// LUA_HEADER_COPY_OFF ///

class Text : public DrawableSF, public sf::String
{
public:
    Text(Group *parent, const std::string &name = "", const std::string &str = "");
    static Text *Create(Group *parent, const std::string &name) {return new Text(parent, name);}

    /// LUA_EXPORT_ON(Text, Drawable) ///

    // Clone & Create
    /** Create a clone of the object and return an handler to the clone. */
    virtual Text *Clone() const {return CloneToGroup(NULL);}

    /** Create a clone of the object, change the parent and return an handler to the clone. */
    virtual Text *CloneToGroup(Group *newParent) const {Text *obj = new Text(*this); obj->RegisterClone(newParent); return obj;}

    // Text
    /** Access text */
    //LUA// luaProperty(std::string, text, GetText, SetText);

    /** Set text */
    void SetText(const std::string& text) {m_originalText = text;InvalidateText();}

    /** Get text */
    std::string GetText() const {return m_originalText;}

    // Size
    /** Return the size of the texte */
    Vector2D GetTextSize();

    // Width
    /** Access the text max width (see SetMaxWidth) */
    //LUA// luaProperty(float, maxWidth, GetMaxWidth, SetMaxWidth);

    /** Set the maximum text width. It means that if a line width is greater than \\a maxWidth, then some line return will be added. \\a maxWidth <= 0 means no limit (default). */
    void SetMaxWidth(float maxWidth) {m_maxWidth = maxWidth; InvalidateText();}

    /** Get the maximum text width */
    float GetMaxWidth() const {return m_maxWidth;}

    /** Get / Set if words can be cut if a maximum width is set. Default is false. */
    //LUA// luaProperty(bool, canCutWords, IsCutWords, SetCutWords);

    /** Set if words can be cut if a maximum width is set. Default is false. */
    void SetCutWords(bool cutWords) {m_canCutWords = cutWords; InvalidateText();}

    /** Return true if words can be cut. Default is false. */
    bool IsCutWords() const {return m_canCutWords;}

    // -------------- From sf::Drawable --------------
    /** Acces object's color. */
    //LUA// luaProperty(Color, color, GetColor, SetColor);
    /** Set object's color. */
    //LUA// void SetColor(const Color& Col);
    /** Get object's color. */
    Color GetColor() const {return sf::Drawable::GetColor();}
    /** Set object's scale */
    //LUA// void SetScale(const Vector2D& Scale);
    /** Get object's scale */
    Vector2D GetScale() const {return sf::Drawable::GetScale();}
    /** Set object's scale (x) */
    //LUA// void SetScaleX(float FactorX);
    /** Set object's scale (y) */
    //LUA// void SetScaleY(float FactorY);
    /** Set object's center */
    //LUA// void SetCenter(const Vector2D& Center);
    /** Get object's center */
    Vector2D GetCenter() const {return sf::Drawable::GetCenter();}

    // From sf::String
    /** Set the Font of the string */
    void SetFont(const Font& charFont) {sf::String::SetFont(charFont); InvalidateText();}
    /** Get the Font of the string */
    //LUA// const Font& GetFont() const;

    /** Set the text size of the string */
    void SetSize(float size) {sf::String::SetSize(size); InvalidateText();}
    /** Get the text size of the string */
    //LUA// float GetSize() const;

    // Style
    /** Acces bold parameter */
    //LUA// luaProperty(bool, bold, SetBold, GetBold);
    /** Set if the Text should be rendered bold */
    void SetBold(bool isBold);
    /** Tells if the Text should be rendered bold */
    bool GetBold() const;

    /** Acces bold parameter */
    //LUA// luaProperty(bool, italic, SetItalic, GetItalic);
    /** Set if the Text should be rendered italic */
    void SetItalic(bool isItalic);
    /** Tells if the Text should be rendered italic */
    bool GetItalic() const;

    /** Acces bold parameter */
    //LUA// luaProperty(bool, underligned, SetUnderligned, GetUnderligned);
    /** Set if the Text should be rendered underligned */
    void SetUnderligned(bool isUnderligned);
    /** Tells if the Text should be rendered underligned */
    bool GetUnderligned() const;
    // /!\ // void SetStyle(unsigned long textStyle) {SetStyle(textStyle); InvalidateText();}
    // /!\ //LUA// unsigned long GetStyle() const;

    /** Get the position of the character \\a Index in the text. */
    //LUA// Vector2D GetCharacterPos(std::size_t Index) const;
    // /!\ //LUA// FloatRect GetRect() const;

    /// LUA_EXPORT_OFF ///

    // No Lua
    // Lua
    //static void RegisterClassLua(lua_State *myLuaState);

    // Acces to position & rotation
    Vector2D GetPosition() const {return DrawableSF::GetPosition();}
    virtual void SetPosition(const Vector2D &pos) {DrawableSF::SetPosition(pos);}
    float GetRotation() const {return DrawableSF::GetRotation();}
    virtual void SetRotation(float angle) {DrawableSF::SetRotation(angle);}
    void Rotate(float angle) {DrawableSF::Rotate(angle);}
    void Move(const Vector2D &mvt) {DrawableSF::Move(mvt);}

    // Arrange text if needed
    virtual void ProceedMvt(float time);

    // Render
    virtual void Render(sf::RenderTarget &App) {RenderSf(App, this);}

    // Acces to sf::Drawable
    ADD_GET_TYPE_FN(OBJ_TYPE_DRAWABLE_TEXT)
protected:
    void InvalidateText() {m_isChanged = true;}

    void ArrangeTextForWidth();

    // Max width
    bool m_isChanged;
    bool m_canCutWords;
    float m_maxWidth;
    std::string m_originalText;

private:
};
//------------------------------------------------------------------------------
} // namespace En

#endif // TEXT_HPP
